/****************************************************************
 *  DISASM.C
 *		Tabellengesteuerter Dis- und Reassembler
 *		========================================
 *		Table controlled dis- and reassembler
 *		=====================================
 *
 *  Copyright 1998 Bodo Wenzel     BodoWenzel@web.de
 *
 *  This source should be compiled with GCC or any other
 *  ANSI-C compiler. Where appropriate use short ints.
 *
 *  Sorry for the poor and german comments ;-)
 *
 *  Historie:
 *  02.01.99	Bodo Wenzel	V1.0, Erstellung erste Version
 ****************************************************************/

#define  VERSION  "V1.0"

#define  EMAIL  "BodoWenzel@web.de"

/* === Einfgedateien ========================================= */

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* === Konstanten ============================================= */

#define  FALSE  (0)
#define  TRUE   (!FALSE)

#define  TEXT_MAX  (100)

#define  NUMBER_MAX   (5+20)
#define  PARA_MAX     (20)
#define  OPCODE_MAX   (20)
#define  OP_PARA_MAX  (4)

#define  OFFS_MAX  (0xFFFFL)
/* muss (2^n)-1 sein */

#define  FIND_MAX   (4)
#define  DEF_LINES  (20)
#define  DUMP_LEN   (16)

#define  LAB_A_MAX  (TEXT_MAX-30)

#define  PARA_PLUS   (-1)
#define  PARA_MINUS  (-2)
#define  PARA_END    (-3)

#define  STACK_MAX  (10)

/* === Globale Typen und Variablen ============================ */

typedef  struct {
  char  name;
  char  *text;
  }  NUMBER;

typedef  struct {
  char   name;
  int    format;
  char   *rpn;
  }  PARA;

typedef  struct {
  int   para[OP_PARA_MAX];
  char  *text;
  }  ENTRY;

typedef  struct {
  char   name;
  ENTRY  entry[256];
  }  OPCODE;

typedef  unsigned char   BYTE;

int  opt_lines;

long  offs_first,offs_last,offs_act;
BYTE  find[FIND_MAX],find_n;

char    *tab_name,*tab_comment;
NUMBER  tab_number[NUMBER_MAX];
enum { LAB_A=0, LAB_B, LAB_W, LAB_S, LAB_C, LAB_END };
PARA    tab_para[PARA_MAX];
ENTRY	tab_other[LAB_END];
OPCODE  tab_opcode[OPCODE_MAX];
int     tab_first;

int  dis_mode,no_opcode;

BYTE  bin_buffer[OFFS_MAX+1];

char  tag_buffer[OFFS_MAX+1];

/* === Funktionsprototypen ==================================== */

int command(char *cmd);

long tab_read(char *filename);

long dis_line(long offset,char *line);
long dis_opcode(int i,long offset,char *line);
long dis_entry(ENTRY *e,long offset,char *line);
void dis_para(int i,long offset,char *line);

long bin_read(char *filename,long offset);

int tag_read(char *filename);
int tag_save(char *filename);

int reassemble(char *filename);

/* === M A I N ================================================ */

int main(int argc, char **argv) {
/*-----------------------------*/
int   i;
char  l[TEXT_MAX];

puts("Welcome to DISASM " VERSION ". Type '?' for help.");

opt_lines=DEF_LINES;
tab_comment=NULL;
offs_act=0;
find_n=0;
dis_mode=LAB_B;
memset(bin_buffer,0xFF,sizeof(bin_buffer));
memset(tag_buffer,' ',sizeof(tag_buffer));
offs_first=0;
offs_last=0;

for (i=1; i<argc; i++) {
  printf("%04lX:%s\n",offs_act,argv[i]);
  if (!command(argv[i]))
    break;
  }

for (;;) {
  printf("%04lX>",offs_act);
  (void)gets(l);
  if (l[0]!='\0')
    (void)command(l);
  }
}

/* === Hilfe und Kommandointerpreter ========================== */

int command(char *cmd) {
/*--------------------*/
long  ret;
long  v;
char  *c;
long  o;
long  n1,n2;
char  l_a_flag,l[TEXT_MAX];

switch (toupper(*cmd)) {

case '?':			/* help				*/
  puts(
"DISASM " VERSION " (" __DATE__ ") by Bodo Wenzel (" EMAIL ")\n"
"This program is freeware - no warranty - all rights reserved\n"
"Thanks to Ward Christensen for REZ as prototype\n"
"?                  help\n"
"Olines             Option: default lines\n"
"Tfile              load Table file\n"
"Bfile[,addr]       load Binary file\n"
"Pfile              load Params file\n"
"F[addr][,vv[..]]   Find byte(s)\n"
"C[addr],type       set Control tag\n"
"                   (First Ascii Byte Word Space Code Last)\n"
"C[addr],K          Kill Control tag\n"
"C*,K               Kill all Control tags\n"
"C[addr][,[end]]    list Control tags\n"
"D[addr][,[end]]    Dump\n"
"L[addr][,[end]]    List\n"
"A[addr][,[end]]    Auto set control tags (L without output)\n"
"Sfile              save Params file\n"
"Gfile              Generate output file\n"
"X                  eXit"
  );
  return(TRUE);

case 'O':			/* Olines			*/
  if (sscanf(++cmd,"%d",&opt_lines)!=1) {
    puts("Wrong number");
    opt_lines=DEF_LINES;
    return(FALSE);
    }
  return(TRUE);

case 'T':			/* Tfile			*/
  ret=tab_read(++cmd);
  if (ret>0)
    printf("Format error in table file, line %ld\n",ret);
  if (ret!=0)
    tab_comment=NULL;
  return(ret==0);

case 'B':			/* Bfile[,addr]			*/
  c=strchr(++cmd,',');
  if (c!=NULL) {
    *c='\0';
    if (sscanf(++c,"%lX",&v)!=1) {
      puts("Missing address");
      return(FALSE);
      }
    offs_act=v&OFFS_MAX;
    }

  ret=bin_read(cmd,offs_act);
  offs_act+=ret;
  offs_act&=OFFS_MAX;
  return(ret!=0);

case 'P':			/* Pfile			*/
  return(tag_read(++cmd));

case 'F':			/* F[addr][,vv[..]]		*/
  c=strchr(++cmd,',');
  if (c!=NULL) {
    *c++='\0';
    find_n=0;
    }

  if (sscanf(cmd,"%lX",&v)!=1)
    offs_act++;
  else
    offs_act=v;
  offs_act&=OFFS_MAX;

  while (c!=NULL) {
    if (find_n==FIND_MAX) {
      printf("Too many bytes (max %u)\n",FIND_MAX);
      return(FALSE);
      }
    if (sscanf(c,"%lX",&v)!=1) {
      puts("Missing byte");
      return(FALSE);
      }
    if (v>0xFF) {
      puts("Byte too big");
      return(FALSE);
      }
    find[find_n++]=(BYTE)v&0xFF;
    c=strchr(c,',');
    if (c!=NULL)
      *c++='\0';
    }

  if (find_n==0) {
    puts("No bytes specified");
    return(FALSE);
    }

  for (;;) {
    if (find[0]==bin_buffer[offs_act]) {
      for (o=offs_act+1, n1=1; n1<find_n; n1++) {
        o&=OFFS_MAX;
        if (find[n1]!=bin_buffer[o])
          break;
        o++;
        }
      if ((BYTE)n1==find_n) {
        printf("Found at %04lX\n",offs_act);
        return(TRUE);
        }
      }
    offs_act++;
    offs_act&=OFFS_MAX;
    if (offs_act==0) {
      puts("Not found");
      return(FALSE);
      }
    }

case 'C':			/* C[addr],type			*/
  c=strchr(++cmd,',');		/* C*,K				*/
  if (c!=NULL)			/* C[addr][,[end]]		*/
    *c++='\0';

  if (*cmd=='*') {
    if (c==NULL) {
      puts("Missing Comma");
      return(FALSE);
      }
    if (toupper(*c)!='K') {
      puts("Only 'K' allowed");
      return(FALSE);
      }
    memset(tag_buffer,' ',sizeof(tag_buffer));
    offs_first=0;
    offs_last=0;
    return(TRUE);
    }

  if (sscanf(cmd,"%lX",&v)==1)
    offs_act=v&OFFS_MAX;

  n1=opt_lines;
  o=0;				/* gegen gcc-Warnung		*/
  if (c!=NULL) {
    if (*(c+1)=='\0')
      switch (toupper(*c)) {
      case 'A':
      case 'B':
      case 'W':
      case 'S':
      case 'C':
        tag_buffer[offs_act]=toupper(*c);
        return(TRUE);
      case 'F':
        offs_first=offs_act;
        return(TRUE);
      case 'L':
        offs_last=offs_act;
        return(TRUE);
      case 'K':
        tag_buffer[offs_act]=' ';
        return(TRUE);
      default:
        break;
        }
    if (sscanf(c,"%lX",&v)==1) {
      o=v&OFFS_MAX;
      if (offs_act<=o)
        n1=0;
      }
    }

  for (;;) {
    if (offs_act==offs_last)
      printf("%04lX,L\n",offs_act);
    if (offs_act==offs_first)
      printf("%04lX,F\n",offs_act);
    if (tag_buffer[offs_act]!=' ') {
      printf("%04lX,%c\n",offs_act,tag_buffer[offs_act]);
      if (n1!=0)
        if (--n1==0)
          break;
      }
    if (n1==0 && offs_act==o)
      break;
    offs_act++;
    offs_act&=OFFS_MAX;
    if (offs_act==0)
      return(TRUE);
    }
  offs_act++;
  offs_act&=OFFS_MAX;
  return(TRUE);

case 'D':			/* D[addr][,[end]]		*/
  c=strchr(++cmd,',');
  if (c!=NULL)
    *c++='\0';

  if (sscanf(cmd,"%lX",&v)==1)
    offs_act=v&OFFS_MAX;

  n1=opt_lines*DUMP_LEN;
  if (c!=NULL)
    if (sscanf(c,"%lX",&v)==1) {
      o=v&OFFS_MAX;
      if (offs_act<=o)
        n1=o-offs_act+1;
      }

  n2=0;
  while (n1--!=0) {
    if (n2==0)
      printf("%04lX  ",offs_act);
    printf("%02X ",bin_buffer[offs_act]);
    offs_act++;
    offs_act&=OFFS_MAX;
    if (++n2==DUMP_LEN) {
      putchar(' ');
      offs_act-=n2;
      offs_act&=OFFS_MAX;
      for (; n2!=0; n2--) {
        if (isprint(bin_buffer[offs_act]&0x7F))
          putchar(bin_buffer[offs_act]&0x7F);
        else
          putchar('.');
        offs_act++;
        offs_act&=OFFS_MAX;
        }
      putchar('\n');
      }
    }

  if (n2!=0) {
    for (n1=n2; n1<DUMP_LEN; n1++)
      printf("   ");
    putchar(' ');
    offs_act-=n2;
    offs_act&=OFFS_MAX;
    for (; n2!=0; n2--) {
      if (isprint(bin_buffer[offs_act]&0x7F))
        putchar(bin_buffer[offs_act]&0x7F);
      else
        putchar('.');
      offs_act++;
      offs_act&=OFFS_MAX;
      }
    putchar('\n');
    }
  return(TRUE);

case 'L':			/* L[addr][,[end]]		*/
case 'A':			/* A[addr][,[end]]		*/
  l_a_flag=toupper(*cmd);

  if (tab_comment==NULL) {
    puts("No table loaded");
    return(FALSE);
    }

  c=strchr(++cmd,',');
  if (c!=NULL)
    *c++='\0';

  if (sscanf(cmd,"%lX",&v)==1)
    offs_act=v&OFFS_MAX;

  n1=opt_lines;
  o=0;				/* gegen gcc-Warnung		*/
  if (c!=NULL) {
    if (sscanf(c,"%lX",&v)==1) {
      o=v&OFFS_MAX;
      if (offs_act<=o)
        n1=0;
      }
    }

  for (;;) {
    n2=offs_act;
    offs_act=dis_line(offs_act,l);
    if (l_a_flag=='L')
      puts(l);
    if (n1!=0) {
      if (--n1==0)
        return(TRUE);
    } else if (offs_act>o || offs_act<=n2)
      return(TRUE);
    }

case 'S':			/* Sfile			*/
  return(tag_save(cmd+1));

case 'G':			/* Gfile			*/
  if (tab_comment==NULL) {
    puts("No table loaded");
    return(FALSE);
    }
  return(reassemble(cmd+1));

case 'X':			/* X				*/
  /* alle "malloc()" werden vom OS freigegeben... */
  puts("Bye.");
  exit(0);

default:
  puts("Unknown command");
  return(FALSE);
  }
}

/* === Tabelle einlesen und aufbauen ========================== */

long tab_read(char *filename) {
/*---------------------------*/
int     i_number,i_para,i_opcode;
int     i,j,k;
FILE    *f;
size_t  l;
char    *p;

tab_comment=NULL;
tab_name="";
for (i=0; i<NUMBER_MAX; i++)
  tab_number[i].name='\0';
i_number=LAB_END-1;
for (i=0; i<PARA_MAX; i++)
  tab_para[i].name='\0';
i_para=-1;
for (i=0; i<LAB_END; i++)
  tab_other[i].text=NULL;
for (i=0; i<OPCODE_MAX; i++)
  tab_opcode[i].name='\0';
i_opcode=-1;

f=fopen(filename,"rb");
if (f==NULL) {
  puts("Can't open table file");
  return(-1);
  }

fseek(f,0L,SEEK_END);
l=(size_t)ftell(f);
fseek(f,0L,SEEK_SET);

p=(char *)malloc((l+1)*sizeof(char));
if (p==NULL) {
  fclose(f);
  puts("Can't allocate memory for table");
  return(-1);
  }

if (fread(p,sizeof(char),l,f)!=l) {
  fclose(f);
  puts("Can't read table file");
  return(-1);
  }
fclose(f);
p[l]='\0';

p=strtok(p,"\r\n");
l=1;
for (;;) {
  if (p==NULL)
    return((long)l);

  switch (toupper(*p)) {

  case '\r':
  case '\n':
  case ';':			/* ;comment			*/
    break;

  case 'T':			/* Ttablename			*/
    tab_name=++p;
    break;

  case 'C':			/* Ccommentformat		*/
    tab_comment=++p;
    if (*p=='\0')
      return((long)l);
    break;

  case 'L':			/* Ll_labelformat		*/
    p++;
    switch (toupper(*p)) {
    case 'A':
      i=LAB_A;
      break;
    case 'B':
      i=LAB_B;
      break;
    case 'W':
      i=LAB_W;
      break;
    case 'S':
      i=LAB_S;
      break;
    case 'C':
      i=LAB_C;
      break;
    default:
      return((long)l);
      }
    if (tab_number[i].name!='\0')
      return((long)l);
    tab_number[i].name=toupper(*p);
    do
      p++;
    while ((*p==' ' || *p=='\t') && *p!='\0');
    tab_number[i].text=p;
    break;

  case 'N':			/* Nn_numberformat		*/
    if (++i_number==NUMBER_MAX)
      return((long)l);
    if (*++p<' ')
      return((long)l);
    tab_number[i_number].name=*p++;
    while ((*p==' ' || *p=='\t') && *p!='\0')
      p++;
    tab_number[i_number].text=p;
    break;

  case 'P':			/* Pp_l/n_rpn			*/
    if (++i_para==PARA_MAX)
      return((long)l);
    if (*++p<' ')
      return((long)l);
    tab_para[i_para].name=*p++;
    while ((*p==' ' || *p=='\t') && *p!='\0')
      p++;
    if (*p<' ')
      return((long)l);
    for (k=0; k<=i_number; k++)
      if (*p==tab_number[k].name) {
        tab_para[i_para].format=k;
        break;
        }
    if (k>i_number)
      return((long)l);
    p++;
    while ((*p==' ' || *p=='\t') && *p!='\0')
      p++;
    tab_para[i_para].rpn=p;
    break;

  case 'D':			/* Dl_p.._:defformat		*/
    p++;
    switch (toupper(*p)) {
    case 'A':
      i=LAB_A;
      break;
    case 'B':
      i=LAB_B;
      break;
    case 'W':
      i=LAB_W;
      break;
    case 'S':
      i=LAB_S;
      break;
    case 'C':
      i=LAB_C;
      break;
    default:
      return((long)l);
      }
    if (tab_other[i].text!=NULL)
      return((long)l);
    p++;
    j=-1;
    while (*p!=':' && *p!='\0') {
      if (*p!=' ' && *p!='\t') {
        if (++j==OP_PARA_MAX)
          return((long)l);
        if (*p=='+')
          tab_other[i].para[j]=PARA_PLUS;
        else if (*p=='-')
          tab_other[i].para[j]=PARA_MINUS;
        else {
          for (k=0; k<=i_para; k++)
            if (*p==tab_para[k].name) {
              tab_other[i].para[j]=k;
              break;
              }
          if (k>i_para)
            return((long)l);
          }
        }
      p++;
      }
    while (++j!=OP_PARA_MAX)
      tab_other[i].para[j]=PARA_END;
    if (*p++=='\0')
      return((long)l);
    tab_other[i].text=p;
    break;

  case 'O':			/* Oo				*/
    if (++i_opcode==OPCODE_MAX)
      return((long)l);
    if (*++p<' ')
      return((long)l);
    tab_opcode[i_opcode].name=*p;
    for (i=0; i<256; i++)
      tab_opcode[i_opcode].entry[i].text=NULL;
    break;

  case ' ':			/* _xx_p.._:disformat		*/
    if (i_opcode<0)
      return((long)l);
    if (sscanf(++p,"%2X",&i)==0)
      return((long)l);
    if (tab_opcode[i_opcode].entry[i].text!=NULL)
      return((long)l);
    p+=2;
    j=-1;
    while (*p!=':' && *p!='\0') {
      if (*p!=' ' && *p!='\t') {
        if (++j==OP_PARA_MAX)
          return((long)l);
        if (*p=='+')
          tab_opcode[i_opcode].entry[i].para[j]=PARA_PLUS;
        else if (*p=='-')
          tab_opcode[i_opcode].entry[i].para[j]=PARA_MINUS;
        else {
          for (k=0; k<=i_para; k++)
            if (*p==tab_para[k].name) {
              tab_opcode[i_opcode].entry[i].para[j]=k;
              break;
              }
          if (k>i_para) {
            for (k=0; k<i_opcode; k++)
              if (*p==tab_opcode[k].name) {
                tab_opcode[i_opcode].entry[i].para[j]=PARA_MAX+k;
                break;
                }
            if (k>=i_opcode)
              return((long)l);
            }
          }
        }
      p++;
      }
    while (++j!=OP_PARA_MAX)
      tab_opcode[i_opcode].entry[i].para[j]=PARA_END;
    if (*p++=='\0')
      return((long)l);
    tab_opcode[i_opcode].entry[i].text=p;
    break;

  case 'E':			/* E				*/
    if (tab_comment==NULL)
      return((long)l);
    for (i=0; i<LAB_END; i++)
      if (tab_number[i].name=='\0')
        return((long)l);
    for (i=0; i<LAB_END; i++)
      if (tab_other[i].text==NULL)
        return((long)l);
    if (i_opcode<0)
      return((long)l);
    tab_first=i_opcode;
    printf("\"%s\" [N%d/P%d/O%d] loaded\n",
      tab_name,++i_number,++i_para,++i_opcode);
    return(0);

  default:
    return((long)l);
    }
  p=strtok(NULL,"\r\n");
  l++;
  }
}

/* === Disassembler =========================================== */

long dis_line(long offset,char *line) {
/*-----------------------------------*/
ENTRY  *e;
long   o,n;
char   l[TEXT_MAX],*t,*p;

switch (tag_buffer[offset]) {
case 'A':
  dis_mode=LAB_A;
  break;
case 'B':
  dis_mode=LAB_B;
  break;
case 'W':
  dis_mode=LAB_W;
  break;
case 'S':
  dis_mode=LAB_S;
  break;
case 'C':
  dis_mode=LAB_C;
  break;
default:
  break;
  }

if (tag_buffer[offset]!=' ') {
  sprintf(line,tab_number[dis_mode].text,offset);
  line+=strlen(line);
  *line++=':';
  }

switch (dis_mode) {
case LAB_A:
  o=offset;
  n=0;
  do {
    if (!isprint(bin_buffer[o]))
      break;
    o++;
    n++;
    if (o>OFFS_MAX || n>=LAB_A_MAX)
      break;
  } while (tag_buffer[o]==' ');

  if (o==offset)
    return(dis_entry(&tab_other[LAB_B],offset,line));

  e=&tab_other[LAB_A];
  for (o=0; o<n; o++)
    l[o]=bin_buffer[offset++];
  l[o]='\0';
  offset&=OFFS_MAX;
  break;
case LAB_W:
  return(dis_entry(&tab_other[LAB_W],offset,line));
case LAB_S:
  o=offset;
  do {
    o++;
    if (o>OFFS_MAX)
      break;
  } while (tag_buffer[o]==' ');

  e=&tab_other[LAB_S];
  sprintf(l,"%lu",o-offset);
  offset=o&OFFS_MAX;
  break;
case LAB_C:
  no_opcode=FALSE;
  o=dis_opcode(tab_first,offset,line);
  if (!no_opcode)
    return(o);
  return(dis_entry(&tab_other[LAB_C],offset,line));
default:			/* default ist auch LAB_B	*/
  return(dis_entry(&tab_other[LAB_B],offset,line));
  }

t=e->text;
do {
  if (*t=='%') {
    if (*++t=='%')
      *line++=*t;
    else if (*t=='1')
      for (p=l; *p!='\0'; p++)
        *line++=*p;
    else
      puts("Index error in disassemble entry");
  } else
    *line++=*t;
} while (*t++!='\0');
return(offset);
}

long dis_opcode(int i,long offset,char *line) {
/*-------------------------------------------*/
ENTRY  *e;

e=&tab_opcode[i].entry[bin_buffer[offset]];

if (e->text==NULL) {
  no_opcode=TRUE;
  *line='\0';
  return(offset);
  }

offset++;
offset&=OFFS_MAX;

return(dis_entry(e,offset,line));
}

long dis_entry(ENTRY *e,long offset,char *line) {
/*---------------------------------------------*/
char  *t,*p;
int   i,p_i;
char  p_l[OP_PARA_MAX][TEXT_MAX];

for (i=0; i<OP_PARA_MAX; i++) {
  p_i=e->para[i];
  if (p_i==PARA_END)
    break;
  if (p_i==PARA_PLUS) {
    offset++;
    offset&=OFFS_MAX;
    p_l[i][0]='\0';
  } else if (p_i==PARA_MINUS) {
    offset--;
    offset&=OFFS_MAX;
    p_l[i][0]='\0';
  } else if (p_i<PARA_MAX) {
    offset++;
    offset&=OFFS_MAX;
    dis_para(p_i,offset,p_l[i]);
  } else
    offset=dis_opcode(p_i-PARA_MAX,offset,p_l[i]);
  }

while (i<OP_PARA_MAX)
  p_l[i++][0]='\0';

t=e->text;
do {
  if (*t=='%') {
    if (*++t=='%')
      *line++=*t;
    else if (*t>='1' && *t<='9')
      for (p=p_l[*t-'1']; *p!='\0'; p++)
        *line++=*p;
    else
      puts("Index error in disassemble entry");
  } else
    *line++=*t;
} while (*t++!='\0');
return(offset);
}

void dis_para(int i,long offset,char *line) {
/*-----------------------------------------*/
long  stack[STACK_MAX];
int   sp;
char  *p;
int   format;

*line='\0';
sp=0;

for (p=tab_para[i].rpn; *p!='\0'; p++)
  switch (*p) {
  case ' ':
  case '\t':
    break;
  case 'u':
    p++;
    if (!isdigit(*p)) {
      printf("Index error in parameter '%c'\n",tab_para[i].name);
      return;
      }
    stack[sp++]=bin_buffer[(offset-*p+'0'-1)&OFFS_MAX];
    break;
  case 's':
    p++;
    if (!isdigit(*p)) {
      printf("Index error in parameter '%c'\n",tab_para[i].name);
      return;
      }
    stack[sp]=bin_buffer[(offset-*p+'0'-1)&OFFS_MAX];
    if (stack[sp]>=0x80)
      stack[sp]-=0x100;
    sp++;
    break;
  case '#':
    p++;
    if (!isxdigit(*p)) {
      printf("Number error in parameter '%c'\n",tab_para[i].name);
      return;
      }
    stack[sp]=0;
    do {
      stack[sp]<<=4;
      stack[sp]|=(*p)<='9' ? (*p)-'0' : toupper(*p)-'A'+10;
      p++;
    } while (isxdigit(*p));
    p--;
    sp++;
    break;
  case '@':
    stack[sp++]=offset;
    break;
  case '+':
    if (sp<2) {
      printf("Stack error in parameter '%c'\n",tab_para[i].name);
      return;
      }
    sp--;
    stack[sp-1]+=stack[sp];
    break;
  case '-':
    if (sp<2) {
      printf("Stack error in parameter '%c'\n",tab_para[i].name);
      return;
      }
    sp--;
    stack[sp-1]-=stack[sp];
    break;
  case '*':
    if (sp<2) {
      printf("Stack error in parameter '%c'\n",tab_para[i].name);
      return;
      }
    sp--;
    stack[sp-1]*=stack[sp];
    break;
  case '/':
    if (sp<2) {
      printf("Stack error in parameter '%c'\n",tab_para[i].name);
      return;
      }
    sp--;
    if (stack[sp]==0) {
      printf("Division by 0 in parameter '%c'\n",
        tab_para[i].name);
      return;
      }
    stack[sp-1]/=stack[sp];
    break;
  case '<':
    if (sp<2) {
      printf("Stack error in parameter '%c'\n",tab_para[i].name);
      return;
      }
    sp--;
    stack[sp-1]<<=stack[sp];
    break;
  case '>':
    if (sp<2) {
      printf("Stack error in parameter '%c'\n",tab_para[i].name);
      return;
      }
    sp--;
    stack[sp-1]>>=stack[sp];
    break;
  case '&':
    if (sp<2) {
      printf("Stack error in parameter '%c'\n",tab_para[i].name);
      return;
      }
    sp--;
    stack[sp-1]&=stack[sp];
    break;
  case '|':
    if (sp<2) {
      printf("Stack error in parameter '%c'\n",tab_para[i].name);
      return;
      }
    sp--;
    stack[sp-1]|=stack[sp];
    break;
  case '^':
    if (sp<2) {
      printf("Stack error in parameter '%c'\n",tab_para[i].name);
      return;
      }
    sp--;
    stack[sp-1]^=stack[sp];
    break;
  default:
    printf("Format error in parameter '%c'\n",tab_para[i].name);
    return;
    }

if (sp!=1) {
  printf("Stack error in parameter '%c'\n",tab_para[i].name);
  return;
  }

format=tab_para[i].format;
if (format<LAB_END)
  switch (tag_buffer[stack[0]&OFFS_MAX]) {
  case 'A':
    format=LAB_A;
    break;
  case 'B':
    format=LAB_B;
    break;
  case 'W':
    format=LAB_W;
    break;
  case 'S':
    format=LAB_S;
    break;
  case 'C':
    format=LAB_C;
    break;
  default:
    tag_buffer[stack[0]&OFFS_MAX]=tab_number[format].name;
    break;
    }

if (strchr(tab_number[format].text,'+')==NULL)
  stack[0]&=OFFS_MAX;
sprintf(line,tab_number[format].text,stack[0]);
}

/* === Binr-Daten ============================================ */

long bin_read(char *filename,long offset) {
/*---------------------------------------*/
FILE    *f;
size_t  l;

f=fopen(filename,"rb");
if (f==NULL) {
  puts("Can't open binary file");
  return(0);
  }

fseek(f,0L,SEEK_END);
l=(size_t)ftell(f);
fseek(f,0L,SEEK_SET);

if (l+offset>OFFS_MAX+1) {
  fclose(f);
  puts("Binary file or offset is too large");
  return(0);
  }

if (fread(bin_buffer+offset,sizeof(char),l,f)!=l) {
  fclose(f);
  puts("Can't read binary file");
  return(0);
  }

fclose(f);
printf("%ld bytes loaded\n",(long)l);
return((long)l);
}

/* === Tag-Daten ============================================== */

int tag_read(char *filename) {
/*--------------------------*/
FILE  *f;
int   t,r;
long  v,n;

f=fopen(filename,"r");
if (f==NULL) {
  puts("Can't open params file");
  return(FALSE);
  }

n=0;
while (!feof(f)) {
  switch (t=fgetc(f)) {
  case EOF:
    continue;
  case 'L':
    r=fscanf(f,"%4lX\n",&v);
    offs_last=v&OFFS_MAX;
    n++;
    break;
  case 'F':
    r=fscanf(f,"%4lX\n",&v);
    offs_first=v&OFFS_MAX;
    n++;
    break;
  case 'A':
  case 'B':
  case 'W':
  case 'S':
  case 'C':
    r=fscanf(f,"%4lX\n",&v);
    tag_buffer[v&OFFS_MAX]=(char)t;
    n++;
    break;
  default:
    r=0;
    break;
    }
  if (r!=1 || ferror(f)) {
    fclose(f);
    puts("Read error on params file");
    return(FALSE);
    }
  }

fclose(f);
printf("%ld control tags loaded\n",n);
return(TRUE);
}

int tag_save(char *filename) {
/*--------------------------*/
FILE  *f;
long  o;
long  n;

f=fopen(filename,"w");
if (f==NULL) {
  puts("Can't open params file");
  return(FALSE);
  }

fprintf(f,"L%04lX\nF%04lX\n",offs_last,offs_first);

n=2;
o=0;
do {
  if (tag_buffer[o]!=' ') {
    n++;
    fprintf(f,"%c%04lX\n",tag_buffer[o],o);
    if (ferror(f))
      break;
    }
  o++;
  o&=OFFS_MAX;
} while (o!=0);

if (ferror(f)) {
  fclose(f);
  puts("Write error on params file");
  return(FALSE);
  }

fclose(f);
printf("%ld control tags written\n",n);
return(TRUE);
}

/* === Reassembler ============================================ */

int reassemble(char *filename) {
/*----------------------------*/
FILE  *f;
long  o,o_old;
long  n;
int   tag;
char  l[TEXT_MAX];

f=fopen(filename,"w");
if (f==NULL) {
  puts("Can't open output file");
  return(FALSE);
  }

n=4;
fprintf(f,tab_comment,"  Source generated by DISASM " VERSION);
fputc('\n',f);
sprintf(l,"  Table %s",tab_name);
fprintf(f,tab_comment,l);
fputc('\n',f);
fputc('\n',f);
fprintf(f,tab_comment,"  Undefined labels:");
fputc('\n',f);

o=0;
do {
  if (o==offs_first && offs_first!=offs_last) {
    o=offs_last;
    continue;
    }
  switch (tag_buffer[o]) {
  case 'A':
    tag=LAB_A;
    break;
  case 'B':
    tag=LAB_B;
    break;
  case 'W':
    tag=LAB_W;
    break;
  case 'S':
    tag=LAB_S;
    break;
  case 'C':
    tag=LAB_C;
    break;
  default:
    tag=-1;
    break;
    }
  if (tag>=0) {
    n++;
    fprintf(f,tab_number[tag].text,o);
    fputc('\n',f);
    if (ferror(f))
      break;
    }
  o++;
  o&=OFFS_MAX;
} while (o!=0);

n+=2;
fputc('\n',f);
fprintf(f,tab_comment,"  Source:");
fputc('\n',f);

if (!ferror(f))
  for (o=offs_first; o<offs_last || offs_last==0; ) {
    o_old=o;
    o=dis_line(o,l);
    n++;
    fputs(l,f);
    fputc('\n',f);
    if (ferror(f))
      break;
    if (o<=o_old)
      break;
    }

n+=2;
fputc('\n',f);
fprintf(f,tab_comment," End.");
fputc('\n',f);

if (ferror(f)) {
  fclose(f);
  puts("Write error on output file");
  return(FALSE);
  }

fclose(f);
printf("%ld lines written\n",n);
return(TRUE);
}

/* === Ende =================================================== */
